@@ -11,6 +11,7 @@ native optimizer, as well as fetch and build ports like zlib and sdl2
 from __future__ import print_function
 import logging
 import os
+import subprocess
 import sys
 
 from tools import shared
@@ -44,14 +45,6 @@ SYSTEM_TASKS = [
     'compiler-rt',
     'emmalloc',
     'emmalloc_debug',
-    'gl',
-    'gl-emu',
-    'gl-emu-webgl2',
-    'gl-mt',
-    'gl-mt-emu',
-    'gl-mt-emu-webgl2',
-    'gl-mt-webgl2',
-    'gl-webgl2',
     'html5',
     'libc',
     'libc++',
@@ -61,6 +54,7 @@ SYSTEM_TASKS = [
     'libc-mt',
     'pthreads',
     'pthreads_stub',
+    'ubsan-minimal-rt-wasm',
 ]
 
 for debug in ['', '_debug']:
@@ -69,16 +63,26 @@ for debug in ['', '_debug']:
       for tracing in ['', '_tracing']:
         SYSTEM_TASKS += ['dlmalloc' + debug + noerrno + threadsafe + tracing]
 
+for mt in ['', '-mt']:
+  for emu in ['', '-emu']:
+    for webgl2 in ['', '-webgl2']:
+      for ofb in ['', '-ofb']:
+        SYSTEM_TASKS += ['gl' + mt + emu + webgl2 + ofb]
+
 USER_TASKS = [
     'binaryen',
     'bullet',
+    'bzip2',
     'cocos2d',
     'freetype',
+    'harfbuzz',
     'icu',
+    'libjpeg',
     'libpng',
     'ogg',
     'regal',
     'sdl2',
+    'sdl2-mt',
     'sdl2-gfx',
     'sdl2-image',
     'sdl2-image-png',
@@ -108,7 +112,7 @@ are running multiple build commands in parallel, confusion can occur).
 
 Usage:
 
-  embuilder.py OPERATION TASK1 [TASK2..] [--lto]
+  embuilder.py OPERATION TASK1 [TASK2..] [--pic] [--lto]
 
 Available operations and tasks:
 
@@ -119,6 +123,7 @@ Issuing 'embuilder.py build ALL' causes each task to be built.
 Flags:
 
   --lto  Build bitcode files, for LTO with the LLVM wasm backend
+  --pic  Build as position independent code (used by MAIN_MODULE/SIDE_MODULE)
 
 It is also possible to build native_optimizer manually by using CMake. To
 do that, run
@@ -132,19 +137,20 @@ and set up the location to the native optimizer in ~/.emscripten
 
 
 def build(src, result_libs, args=[]):
-  if shared.Settings.WASM_BACKEND:
-    args = args + ['-s', 'WASM_OBJECT_FILES=%d' % shared.Settings.WASM_OBJECT_FILES]
+  if not shared.Settings.WASM_OBJECT_FILES:
+    args += ['-s', 'WASM_OBJECT_FILES=0']
+  if shared.Settings.RELOCATABLE:
+    args += ['-s', 'RELOCATABLE']
   # build in order to generate the libraries
   # do it all in a temp dir where everything will be cleaned up
   temp_dir = temp_files.get_dir()
   cpp = os.path.join(temp_dir, 'src.cpp')
   open(cpp, 'w').write(src)
   temp_js = os.path.join(temp_dir, 'out.js')
-  shared.Building.emcc(cpp, args, output_filename=temp_js)
-
-  # verify
-  if not os.path.exists(temp_js):
-    shared.exit_with_error('failed to build file')
+  try:
+    shared.Building.emcc(cpp, args, output_filename=temp_js)
+  except subprocess.CalledProcessError as e:
+    shared.exit_with_error("embuilder: emcc command failed with %d: '%s'", e.returncode, ' '.join(e.cmd))
 
   for lib in result_libs:
     if not os.path.exists(shared.Cache.get_path(lib)):
@@ -152,14 +158,7 @@ def build(src, result_libs, args=[]):
 
 
 def build_port(port_name, lib_name, params):
-  build(C_BARE, [os.path.join('ports-builds', port_name, lib_name)] if lib_name else [], params)
-
-
-def static_library_name(name):
-  if shared.Settings.WASM_BACKEND and shared.Settings.WASM_OBJECT_FILES:
-    return name + '.a'
-  else:
-    return name + '.bc'
+  build(C_BARE, [lib_name] if lib_name else [], params)
 
 
 def main():
@@ -183,19 +182,25 @@ def main():
       arg = arg[2:]
       if arg == 'lto':
         shared.Settings.WASM_OBJECT_FILES = 0
-        # Reconfigure the cache dir to reflect the change
-        shared.reconfigure_cache()
+      elif arg == 'pic':
+        shared.Settings.RELOCATABLE = 1
+      # Reconfigure the cache dir to reflect the change
+      shared.reconfigure_cache()
 
   args = [a for a in args if not is_flag(a)]
 
   # process tasks
+  libname = shared.static_library_name
 
   auto_tasks = False
   tasks = args
   if 'SYSTEM' in tasks:
     tasks = SYSTEM_TASKS
     auto_tasks = True
-  if 'ALL' in tasks:
+  elif 'USER' in tasks:
+    tasks = USER_TASKS
+    auto_tasks = True
+  elif 'ALL' in tasks:
     tasks = SYSTEM_TASKS + USER_TASKS
     auto_tasks = True
   if auto_tasks:
@@ -225,20 +230,20 @@ def main():
         }
       ''', ['libcompiler_rt.a'])
     elif what == 'libc':
-      build(C_WITH_MALLOC, [static_library_name('libc')])
+      build(C_WITH_MALLOC, [libname('libc')])
     elif what == 'libc-extras':
       build('''
         extern char **environ;
         int main() {
           return (int)environ;
         }
-      ''', [static_library_name('libc-extras')])
+      ''', [libname('libc-extras')])
     elif what == 'struct_info':
       build(C_BARE, ['generated_struct_info.json'])
     elif what == 'emmalloc':
-      build(C_WITH_MALLOC, [static_library_name('libemmalloc')], ['-s', 'MALLOC="emmalloc"'])
+      build(C_WITH_MALLOC, [libname('libemmalloc')], ['-s', 'MALLOC="emmalloc"'])
     elif what == 'emmalloc_debug':
-      build(C_WITH_MALLOC, [static_library_name('libemmalloc_debug')], ['-s', 'MALLOC="emmalloc"', '-g'])
+      build(C_WITH_MALLOC, [libname('libemmalloc_debug')], ['-s', 'MALLOC="emmalloc"', '-g'])
     elif what.startswith('dlmalloc'):
       cmd = ['-s', 'MALLOC="dlmalloc"']
       if '_debug' in what:
@@ -249,11 +254,11 @@ def main():
         cmd += ['-s', 'USE_PTHREADS=1']
       if '_tracing' in what:
         cmd += ['-s', 'EMSCRIPTEN_TRACING=1']
-      build(C_WITH_MALLOC, [static_library_name('lib' + what)], cmd)
+      build(C_WITH_MALLOC, [libname('lib' + what)], cmd)
     elif what in ('libc-mt', 'pthreads'):
-      build(C_WITH_MALLOC, [static_library_name('libc-mt'), static_library_name('libpthreads')], ['-s', 'USE_PTHREADS=1'])
+      build(C_WITH_MALLOC, [libname('libc-mt'), libname('libpthreads')], ['-s', 'USE_PTHREADS=1'])
     elif what == 'libc-wasm':
-      build(C_WITH_STDLIB, [static_library_name('libc-wasm')], ['-s', 'WASM=1'])
+      build(C_WITH_STDLIB, [libname('libc-wasm')], ['-s', 'WASM=1'])
     elif what == 'libc++':
       build(CXX_WITH_STDLIB, ['libc++.a'], ['-s', 'DISABLE_EXCEPTION_CATCHING=0'])
     elif what == 'libc++_noexcept':
@@ -267,7 +272,7 @@ def main():
           y->a();
           return y->y;
         }
-      ''', [static_library_name('libc++abi')])
+      ''', [libname('libc++abi')])
     elif what == 'gl' or what.startswith('gl-'):
       opts = []
       if '-mt' in what:
@@ -276,12 +281,14 @@ def main():
         opts += ['-s', 'LEGACY_GL_EMULATION=1']
       if '-webgl2' in what:
         opts += ['-s', 'USE_WEBGL2=1']
+      if '-ofb' in what:
+        opts += ['-s', 'OFFSCREEN_FRAMEBUFFER=1']
       build('''
         extern "C" { extern void* emscripten_GetProcAddress(const char *x); }
         int main() {
           return int(emscripten_GetProcAddress("waka waka"));
         }
-      ''', [static_library_name('lib' + what)], opts)
+      ''', [libname('lib' + what)], opts)
     elif what == 'native_optimizer':
       build(C_BARE, ['optimizer.2.exe'], ['-O2', '-s', 'WASM=0'])
     elif what == 'compiler_rt_wasm':
@@ -297,7 +304,7 @@ def main():
           return emscripten_compute_dom_pk_code(NULL);
         }
 
-      ''', [static_library_name('libhtml5')])
+      ''', [libname('libhtml5')])
     elif what == 'pthreads_stub':
       build('''
         #include <emscripten/threading.h>
@@ -305,7 +312,11 @@ def main():
           return emscripten_is_main_runtime_thread();
         }
 
-      ''', [static_library_name('libpthreads_stub')])
+      ''', [libname('libpthreads_stub')])
+    elif what == 'ubsan-minimal-rt-wasm':
+      if not shared.Settings.WASM_BACKEND:
+        continue
+      build(C_BARE, ['libubsan_minimal_rt_wasm.a'], ['-fsanitize=undefined', '-fsanitize-minimal-runtime', '-s', 'WASM=1'])
     elif what == 'al':
       build('''
         #include "AL/al.h"
@@ -313,43 +324,49 @@ def main():
           alGetProcAddress(0);
           return 0;
         }
-      ''', [static_library_name('libal')])
+      ''', [libname('libal')])
     elif what == 'icu':
-      build_port('icu', 'libicuuc.bc', ['-s', 'USE_ICU=1'])
+      build_port('icu', libname('libicuuc'), ['-s', 'USE_ICU=1'])
     elif what == 'zlib':
       build_port('zlib', 'libz.a', ['-s', 'USE_ZLIB=1'])
+    elif what == 'bzip2':
+      build_port('bzip2', 'libbz2.a', ['-s', 'USE_BZIP2=1'])
     elif what == 'bullet':
-      build_port('bullet', 'libbullet.bc', ['-s', 'USE_BULLET=1'])
+      build_port('bullet', libname('libbullet'), ['-s', 'USE_BULLET=1'])
     elif what == 'vorbis':
-      build_port('vorbis', 'libvorbis.bc', ['-s', 'USE_VORBIS=1'])
+      build_port('vorbis', libname('libvorbis'), ['-s', 'USE_VORBIS=1'])
     elif what == 'ogg':
-      build_port('ogg', 'libogg.bc', ['-s', 'USE_OGG=1'])
+      build_port('ogg', libname('libogg'), ['-s', 'USE_OGG=1'])
+    elif what == 'libjpeg':
+      build_port('libjpeg', libname('libjpeg'), ['-s', 'USE_LIBJPEG=1'])
     elif what == 'libpng':
-      build_port('libpng', 'libpng.bc', ['-s', 'USE_ZLIB=1', '-s', 'USE_LIBPNG=1'])
+      build_port('libpng', libname('libpng'), ['-s', 'USE_ZLIB=1', '-s', 'USE_LIBPNG=1'])
     elif what == 'sdl2':
-      build_port('sdl2', 'libsdl2.bc', ['-s', 'USE_SDL=2'])
+      build_port('sdl2', libname('libSDL2'), ['-s', 'USE_SDL=2'])
+    elif what == 'sdl2-mt':
+      build_port('sdl2', libname('libSDL2-mt'), ['-s', 'USE_SDL=2', '-s', 'USE_PTHREADS=1'])
     elif what == 'sdl2-gfx':
-      build_port('sdl2-gfx', 'libsdl2_gfx.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'USE_SDL_GFX=2'])
+      build_port('sdl2-gfx', libname('libSDL2_gfx'), ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'USE_SDL_GFX=2'])
     elif what == 'sdl2-image':
-      build_port('sdl2-image', 'libsdl2_image.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2'])
+      build_port('sdl2-image', libname('libSDL2_image'), ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2'])
     elif what == 'sdl2-image-png':
-      build_port('sdl2-image', 'libsdl2_image.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'SDL2_IMAGE_FORMATS=["png"]'])
+      build_port('sdl2-image', libname('libSDL2_image'), ['-s', 'USE_SDL=2', '-s', 'USE_SDL_IMAGE=2', '-s', 'SDL2_IMAGE_FORMATS=["png"]'])
     elif what == 'sdl2-net':
-      build_port('sdl2-net', 'libsdl2_net.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_NET=2'])
+      build_port('sdl2-net', libname('libSDL2_net'), ['-s', 'USE_SDL=2', '-s', 'USE_SDL_NET=2'])
     elif what == 'sdl2-mixer':
-      build_port('sdl2-mixer', 'libsdl2_mixer.a', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_MIXER=2', '-s', 'USE_VORBIS=1'])
+      build_port('sdl2-mixer', 'libSDL2_mixer.a', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_MIXER=2', '-s', 'USE_VORBIS=1'])
     elif what == 'freetype':
       build_port('freetype', 'libfreetype.a', ['-s', 'USE_FREETYPE=1'])
     elif what == 'harfbuzz':
       build_port('harfbuzz', 'libharfbuzz.a', ['-s', 'USE_HARFBUZZ=1'])
     elif what == 'sdl2-ttf':
-      build_port('sdl2-ttf', 'libsdl2_ttf.bc', ['-s', 'USE_SDL=2', '-s', 'USE_SDL_TTF=2', '-s', 'USE_FREETYPE=1'])
+      build_port('sdl2-ttf', libname('libSDL2_ttf'), ['-s', 'USE_SDL=2', '-s', 'USE_SDL_TTF=2', '-s', 'USE_FREETYPE=1'])
     elif what == 'binaryen':
       build_port('binaryen', None, ['-s', 'WASM=1'])
     elif what == 'cocos2d':
-      build_port('cocos2d', None, ['-s', 'USE_COCOS2D=3', '-s', 'USE_ZLIB=1', '-s', 'USE_LIBPNG=1', '-s', 'ERROR_ON_UNDEFINED_SYMBOLS=0'])
+      build_port('cocos2d', libname('libcocos2d'), ['-s', 'USE_COCOS2D=3', '-s', 'USE_ZLIB=1', '-s', 'USE_LIBPNG=1', '-s', 'ERROR_ON_UNDEFINED_SYMBOLS=0'])
     elif what == 'regal':
-      build_port('regal', 'libregal.bc', ['-s', 'USE_REGAL=1'])
+      build_port('regal', libname('libregal'), ['-s', 'USE_REGAL=1'])
     else:
       logger.error('unfamiliar build target: ' + what)
       return 1
