@@ -1,14 +1,15 @@
-#ifdef __EMSCRIPTEN_PTHREADS__
-
 #include <emscripten/threading.h>
 #include <emscripten.h>
 #include <string.h>
 #include <stdlib.h>
+#include <sys/types.h>
 
 #include "webgl1.h"
 #include "webgl1_ext.h"
 #include "webgl2.h"
 
+#if defined(__EMSCRIPTEN_PTHREADS__) && defined(__EMSCRIPTEN_OFFSCREEN_FRAMEBUFFER__)
+
 extern EMSCRIPTEN_WEBGL_CONTEXT_HANDLE emscripten_webgl_do_create_context(const char *target, const EmscriptenWebGLContextAttributes *attributes);
 extern EMSCRIPTEN_RESULT emscripten_webgl_make_context_current_calling_thread(EMSCRIPTEN_WEBGL_CONTEXT_HANDLE context);
 extern EMSCRIPTEN_RESULT emscripten_webgl_do_commit_frame(void);
@@ -624,4 +625,225 @@ RET_SYNC_GL_FUNCTION_3(EM_FUNC_SIG_VIII, void, glGetQueryObjectuivEXT, GLenum, G
 RET_SYNC_GL_FUNCTION_3(EM_FUNC_SIG_VIII, void, glGetQueryObjecti64vEXT, GLenum, GLenum, GLint64 *);
 RET_SYNC_GL_FUNCTION_3(EM_FUNC_SIG_VIII, void, glGetQueryObjectui64vEXT, GLenum, GLenum, GLuint64 *);
 
-#endif // ~__EMSCRIPTEN_PTHREADS__
+#endif // ~(__EMSCRIPTEN_PTHREADS__ && __EMSCRIPTEN_OFFSCREEN_FRAMEBUFFER__)
+
+// Returns a function pointer to the given WebGL 1 extension function, when queried without
+// a GL extension suffix such as "EXT", "OES", or "ANGLE". This function is used by
+// emscripten_GetProcAddress() to implement legacy GL emulation semantics for portability.
+void *_webgl1_match_ext_proc_address_without_suffix(const char *name)
+{
+  RETURN_FN_WITH_SUFFIX(glGenQueries, EXT);
+  RETURN_FN_WITH_SUFFIX(glDeleteQueries, EXT);
+  RETURN_FN_WITH_SUFFIX(glIsQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glBeginQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glEndQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glQueryCounter, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectuiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjecti64v, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectui64v, EXT);
+
+  // WebGL 1 , Extensions
+  RETURN_FN_WITH_SUFFIX(glBindVertexArray, OES);
+  RETURN_FN_WITH_SUFFIX(glDeleteVertexArrays, OES);
+  RETURN_FN_WITH_SUFFIX(glGenVertexArrays, OES);
+  RETURN_FN_WITH_SUFFIX(glIsVertexArray, OES);
+  RETURN_FN_WITH_SUFFIX(glDrawBuffers, WEBGL);
+  RETURN_FN_WITH_SUFFIX(glDrawArraysInstanced, ANGLE);
+  RETURN_FN_WITH_SUFFIX(glDrawElementsInstanced, ANGLE);
+  RETURN_FN_WITH_SUFFIX(glVertexAttribDivisor, ANGLE);
+  RETURN_FN_WITH_SUFFIX(glGenQueries, EXT);
+  RETURN_FN_WITH_SUFFIX(glDeleteQueries, EXT);
+  RETURN_FN_WITH_SUFFIX(glIsQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glBeginQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glEndQuery, EXT);
+  RETURN_FN_WITH_SUFFIX(glQueryCounter, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectuiv, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjecti64v, EXT);
+  RETURN_FN_WITH_SUFFIX(glGetQueryObjectui64v, EXT);
+
+  return 0;
+}
+
+void *emscripten_webgl1_get_proc_address(const char *name)
+{
+  RETURN_FN(glActiveTexture);
+  RETURN_FN(glAttachShader);
+  RETURN_FN(glBindAttribLocation);
+  RETURN_FN(glBindBuffer);
+  RETURN_FN(glBindFramebuffer);
+  RETURN_FN(glBindRenderbuffer);
+  RETURN_FN(glBindTexture);
+  RETURN_FN(glBlendColor);
+  RETURN_FN(glBlendEquation);
+  RETURN_FN(glBlendEquationSeparate);
+  RETURN_FN(glBlendFunc);
+  RETURN_FN(glBlendFuncSeparate);
+  RETURN_FN(glBufferData);
+  RETURN_FN(glBufferSubData);
+  RETURN_FN(glCheckFramebufferStatus);
+  RETURN_FN(glClear);
+  RETURN_FN(glClearColor);
+  RETURN_FN(glClearDepthf);
+  RETURN_FN(glClearStencil);
+  RETURN_FN(glColorMask);
+  RETURN_FN(glCompileShader);
+  RETURN_FN(glCompressedTexImage2D);
+  RETURN_FN(glCompressedTexSubImage2D);
+  RETURN_FN(glCopyTexImage2D);
+  RETURN_FN(glCopyTexSubImage2D);
+  RETURN_FN(glCreateProgram);
+  RETURN_FN(glCreateShader);
+  RETURN_FN(glCullFace);
+  RETURN_FN(glDeleteBuffers);
+  RETURN_FN(glDeleteFramebuffers);
+  RETURN_FN(glDeleteProgram);
+  RETURN_FN(glDeleteRenderbuffers);
+  RETURN_FN(glDeleteShader);
+  RETURN_FN(glDeleteTextures);
+  RETURN_FN(glDepthFunc);
+  RETURN_FN(glDepthMask);
+  RETURN_FN(glDepthRangef);
+  RETURN_FN(glDetachShader);
+  RETURN_FN(glDisable);
+  RETURN_FN(glDisableVertexAttribArray);
+  RETURN_FN(glDrawArrays);
+  RETURN_FN(glDrawElements);
+  RETURN_FN(glEnable);
+  RETURN_FN(glEnableVertexAttribArray);
+  RETURN_FN(glFinish);
+  RETURN_FN(glFlush);
+  RETURN_FN(glFramebufferRenderbuffer);
+  RETURN_FN(glFramebufferTexture2D);
+  RETURN_FN(glFrontFace);
+  RETURN_FN(glGenBuffers);
+  RETURN_FN(glGenerateMipmap);
+  RETURN_FN(glGenFramebuffers);
+  RETURN_FN(glGenRenderbuffers);
+  RETURN_FN(glGenTextures);
+  RETURN_FN(glGetActiveAttrib);
+  RETURN_FN(glGetActiveUniform);
+  RETURN_FN(glGetAttachedShaders);
+  RETURN_FN(glGetAttribLocation);
+  RETURN_FN(glGetBooleanv);
+  RETURN_FN(glGetBufferParameteriv);
+  RETURN_FN(glGetError);
+  RETURN_FN(glGetFloatv);
+  RETURN_FN(glGetFramebufferAttachmentParameteriv);
+  RETURN_FN(glGetIntegerv);
+  RETURN_FN(glGetProgramiv);
+  RETURN_FN(glGetProgramInfoLog);
+  RETURN_FN(glGetRenderbufferParameteriv);
+  RETURN_FN(glGetShaderiv);
+  RETURN_FN(glGetShaderInfoLog);
+  RETURN_FN(glGetShaderPrecisionFormat);
+  RETURN_FN(glGetShaderSource);
+  RETURN_FN(glGetString);
+  RETURN_FN(glGetTexParameterfv);
+  RETURN_FN(glGetTexParameteriv);
+  RETURN_FN(glGetUniformfv);
+  RETURN_FN(glGetUniformiv);
+  RETURN_FN(glGetUniformLocation);
+  RETURN_FN(glGetVertexAttribfv);
+  RETURN_FN(glGetVertexAttribiv);
+  RETURN_FN(glGetVertexAttribPointerv);
+  RETURN_FN(glHint);
+  RETURN_FN(glIsBuffer);
+  RETURN_FN(glIsEnabled);
+  RETURN_FN(glIsFramebuffer);
+  RETURN_FN(glIsProgram);
+  RETURN_FN(glIsRenderbuffer);
+  RETURN_FN(glIsShader);
+  RETURN_FN(glIsTexture);
+  RETURN_FN(glLineWidth);
+  RETURN_FN(glLinkProgram);
+  RETURN_FN(glPixelStorei);
+  RETURN_FN(glPolygonOffset);
+  RETURN_FN(glReadPixels);
+  RETURN_FN(glReleaseShaderCompiler);
+  RETURN_FN(glRenderbufferStorage);
+  RETURN_FN(glSampleCoverage);
+  RETURN_FN(glScissor);
+  RETURN_FN(glShaderBinary);
+  RETURN_FN(glShaderSource);
+  RETURN_FN(glStencilFunc);
+  RETURN_FN(glStencilFuncSeparate);
+  RETURN_FN(glStencilMask);
+  RETURN_FN(glStencilMaskSeparate);
+  RETURN_FN(glStencilOp);
+  RETURN_FN(glStencilOpSeparate);
+  RETURN_FN(glTexImage2D);
+  RETURN_FN(glTexParameterf);
+  RETURN_FN(glTexParameterfv);
+  RETURN_FN(glTexParameteri);
+  RETURN_FN(glTexParameteriv);
+  RETURN_FN(glTexSubImage2D);
+  RETURN_FN(glUniform1f);
+  RETURN_FN(glUniform1fv);
+  RETURN_FN(glUniform1i);
+  RETURN_FN(glUniform1iv);
+  RETURN_FN(glUniform2f);
+  RETURN_FN(glUniform2fv);
+  RETURN_FN(glUniform2i);
+  RETURN_FN(glUniform2iv);
+  RETURN_FN(glUniform3f);
+  RETURN_FN(glUniform3fv);
+  RETURN_FN(glUniform3i);
+  RETURN_FN(glUniform3iv);
+  RETURN_FN(glUniform4f);
+  RETURN_FN(glUniform4fv);
+  RETURN_FN(glUniform4i);
+  RETURN_FN(glUniform4iv);
+  RETURN_FN(glUniformMatrix2fv);
+  RETURN_FN(glUniformMatrix3fv);
+  RETURN_FN(glUniformMatrix4fv);
+  RETURN_FN(glUseProgram);
+  RETURN_FN(glValidateProgram);
+  RETURN_FN(glVertexAttrib1f);
+  RETURN_FN(glVertexAttrib1fv);
+  RETURN_FN(glVertexAttrib2f);
+  RETURN_FN(glVertexAttrib2fv);
+  RETURN_FN(glVertexAttrib3f);
+  RETURN_FN(glVertexAttrib3fv);
+  RETURN_FN(glVertexAttrib4f);
+  RETURN_FN(glVertexAttrib4fv);
+  RETURN_FN(glVertexAttribPointer);
+  RETURN_FN(glViewport);
+  RETURN_FN(glGenQueriesEXT);
+  RETURN_FN(glDeleteQueriesEXT);
+  RETURN_FN(glIsQueryEXT);
+  RETURN_FN(glBeginQueryEXT);
+  RETURN_FN(glEndQueryEXT);
+  RETURN_FN(glQueryCounterEXT);
+  RETURN_FN(glGetQueryivEXT);
+  RETURN_FN(glGetQueryObjectivEXT);
+  RETURN_FN(glGetQueryObjectuivEXT);
+  RETURN_FN(glGetQueryObjecti64vEXT);
+  RETURN_FN(glGetQueryObjectui64vEXT);
+
+  // WebGL 1 Extensions
+  RETURN_FN(glBindVertexArrayOES);
+  RETURN_FN(glDeleteVertexArraysOES);
+  RETURN_FN(glGenVertexArraysOES);
+  RETURN_FN(glIsVertexArrayOES);
+  RETURN_FN(glDrawBuffersWEBGL);
+  RETURN_FN(glDrawArraysInstancedANGLE);
+  RETURN_FN(glDrawElementsInstancedANGLE);
+  RETURN_FN(glVertexAttribDivisorANGLE);
+  RETURN_FN(glGenQueriesEXT);
+  RETURN_FN(glDeleteQueriesEXT);
+  RETURN_FN(glIsQueryEXT);
+  RETURN_FN(glBeginQueryEXT);
+  RETURN_FN(glEndQueryEXT);
+  RETURN_FN(glQueryCounterEXT);
+  RETURN_FN(glGetQueryivEXT);
+  RETURN_FN(glGetQueryObjectivEXT);
+  RETURN_FN(glGetQueryObjectuivEXT);
+  RETURN_FN(glGetQueryObjecti64vEXT);
+  RETURN_FN(glGetQueryObjectui64vEXT);
+
+  return 0;
+}
